import ast

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.python.global_import_datetime import GlobalImportDatetimeRule


def test_global_import_datetime():
    source_tree = ast.parse("""import datetime
my_date = datetime.datetime(2024, 9, 19)
    """)

    occurrences = GlobalImportDatetimeRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_nested_import_datetime():
    source_tree = ast.parse("""from datetime import datetime
my_datetime = datetime(2024, 9, 19)
    """)

    occurrences = GlobalImportDatetimeRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=GlobalImportDatetimeRule.RULE_ID,
        rule_label=GlobalImportDatetimeRule.RULE_LABEL,
        identifier=None,
    )


def test_nested_import_date():
    source_tree = ast.parse("""from datetime import date
my_date = date(2024, 9, 19)
    """)

    occurrences = GlobalImportDatetimeRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=GlobalImportDatetimeRule.RULE_ID,
        rule_label=GlobalImportDatetimeRule.RULE_LABEL,
        identifier=None,
    )


def test_nested_import_renamed():
    source_tree = ast.parse("""from datetime import date as dt
my_date = date(2024, 9, 19)
    """)

    occurrences = GlobalImportDatetimeRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=GlobalImportDatetimeRule.RULE_ID,
        rule_label=GlobalImportDatetimeRule.RULE_LABEL,
        identifier=None,
    )


def test_nested_import_other_things_ok():
    source_tree = ast.parse("""from datetime import UTC""")

    occurrences = GlobalImportDatetimeRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0



from _ast import AST

from boa_restrictor.projections.occurrence import Occurrence

PYTHON_LINTING_RULE_PREFIX = "PBR"
DJANGO_LINTING_RULE_PREFIX = "DBR"


class Rule:
    RULE_ID: str
    RULE_LABEL: str

    filename: str
    source_tree: AST

    @classmethod
    def run_check(cls, *, filename: str, source_tree: AST) -> list[Occurrence]:
        instance = cls(filename=filename, source_tree=source_tree)
        return instance.check()

    def __init__(self, *, filename: str, source_tree: AST):
        """
        A rule is called via pre-commit for a specific file.
        Variable `source_code` is the content of the given file.
        """
        super().__init__()

        self.filename = filename
        self.source_tree = source_tree

    def check(self) -> list[Occurrence]:
        raise NotImplementedError



import argparse
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from boa_restrictor.cli.configuration import is_rule_excluded, is_rule_excluded_per_file, load_configuration
from boa_restrictor.cli.utils import parse_source_code_or_fail
from boa_restrictor.common.noqa import get_noqa_comments
from boa_restrictor.rules import get_rules


def main(argv: Optional[Sequence[str]] = None):
    parser = argparse.ArgumentParser(
        prog="boa-restrictor",
    )
    parser.add_argument(
        "filenames",
        nargs="*",
        help="Filenames to process.",
    )
    parser.add_argument(
        "--config",
        default="pyproject.toml",
        type=str,
        help="Location of pyproject.toml configuration file",
    )
    args = parser.parse_args(argv)

    # Get excluded linting rules from configuration
    configuration = load_configuration(file_path=args.config)
    globally_excluded_rules = configuration.get("exclude", [])
    enable_django_rules = configuration.get("enable_django_rules", True)
    per_file_excluded_rules: dict[str, list[str]] = load_configuration(file_path=args.config).get(
        "per-file-excludes", {}
    )

    # Iterate over all filenames coming from pre-commit...
    occurrences = []
    for filename in args.filenames[1:]:
        # Read source code
        with open(filename) as f:
            source_code = f.read()

        # Parse code through abstract syntax tree
        source_tree = parse_source_code_or_fail(filename=filename, source_code=source_code)

        # Fetch all ignored line comments
        noqa_tokens = get_noqa_comments(source_code=source_code)

        # Iterate over all linters...
        enabled_rules = get_rules(use_django_rules=enable_django_rules)
        for rule_class in enabled_rules:
            # Skip linters, which have been excluded globally via the configuration
            if is_rule_excluded(rule_class=rule_class, excluded_rules=globally_excluded_rules):
                continue

            # Iterate per-file rule exclusions
            if is_rule_excluded_per_file(
                filename=filename, rule_class=rule_class, per_file_excluded_rules=per_file_excluded_rules
            ):
                continue

            # Ensure that line exclusions are respected
            excluded_lines = {token[0] for token in noqa_tokens if rule_class.RULE_ID in token[1]}

            # Add issues to our occurrence list
            occurrences.extend(
                [
                    possible_occurrence
                    for possible_occurrence in rule_class.run_check(filename=filename, source_tree=source_tree)
                    if possible_occurrence.line_number not in excluded_lines
                ]
            )

    # If we have any matches...
    if any(occurrences):
        current_path = Path.cwd()

        # Iterate over them and print details for the user
        for occurrence in occurrences:
            sys.stdout.write(
                f'"{current_path / occurrence.filename}:{occurrence.line_number}": '
                f"({occurrence.rule_id}) {occurrence.rule_label}\n"
            )

    # Since pre-commit will run this function x times, we skip any success or result count messages.

    return bool(any(occurrences))



import argparse
import ast
import os
import sys
from unittest import mock

import pytest

from boa_restrictor.cli.main import main
from boa_restrictor.common.rule import Rule
from boa_restrictor.exceptions.syntax_errors import BoaRestrictorParsingError
from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules import BOA_RESTRICTOR_RULES, DJANGO_BOA_RULES, AsteriskRequiredRule


@mock.patch.object(argparse.ArgumentParser, "parse_args")
def test_main_arguments_parsed(mocked_parse_args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_parse_args.assert_called_once_with(
        ("boa-restrictor", os.path.abspath(sys.argv[0]), "--config", "pyproject.toml"),
    )


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"exclude": ["PBR001"]})
@mock.patch.object(Rule, "run_check")
@mock.patch.object(AsteriskRequiredRule, "run_check")
def test_main_exclude_config_active(mocked_run_checks_asterisk, mocked_rule_run_checks, *args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_run_checks_asterisk.assert_not_called()
    assert mocked_rule_run_checks.call_count == len(BOA_RESTRICTOR_RULES) + len(DJANGO_BOA_RULES) - 1, (
        "We expect all but one rule to be called."
    )


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"per-file-excludes": {"*.py": ["PBR001"]}})
@mock.patch.object(Rule, "run_check")
@mock.patch.object(AsteriskRequiredRule, "run_check")
def test_main_per_file_exclude_config_active(mocked_run_checks_asterisk, mocked_rule_run_checks, *args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_run_checks_asterisk.assert_not_called()
    assert mocked_rule_run_checks.call_count == len(BOA_RESTRICTOR_RULES) + len(DJANGO_BOA_RULES) - 1, (
        "We expect all but one rule to be called."
    )


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={})
@mock.patch("boa_restrictor.cli.main.get_rules")
def test_main_django_rules_default_enabled(mocked_get_rule, *args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_rule.assert_called_with(use_django_rules=True)


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"enable_django_rules": True})
@mock.patch("boa_restrictor.cli.main.get_rules")
def test_main_django_rules_enabled(mocked_get_rule, *args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_rule.assert_called_with(use_django_rules=True)


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"enable_django_rules": False})
@mock.patch("boa_restrictor.cli.main.get_rules")
def test_main_django_rules_disabled(mocked_get_rule, *args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_rule.assert_called_with(use_django_rules=False)


@mock.patch("boa_restrictor.cli.main.get_noqa_comments", return_value=[])
def test_main_noqa_comments_called(mocked_get_noqa_comments):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_noqa_comments.assert_called_once()


@mock.patch.object(ast, "parse", side_effect=SyntaxError)
def test_main_invalid_syntax(*args):
    with pytest.raises(BoaRestrictorParsingError, match=r"Source code of file"):
        main(
            argv=(
                "boa-restrictor",
                os.path.abspath(sys.argv[0]),
                "--config",
                "pyproject.toml",
            )
        )


@mock.patch.object(sys.stdout, "write")
def test_main_occurrences_are_written_to_cli(mocked_write):
    occurrence = Occurrence(
        rule_id="PBR000",
        rule_label="One to rule them all.",
        filename="my_file.py",
        line_number=42,
        identifier="my_function",
    )

    with mock.patch.object(
        Rule,
        "run_check",
        return_value=[occurrence],
    ) as mocked_run_checks:
        main(
            argv=(
                "boa-restrictor",
                os.path.abspath(sys.argv[0]),
                "--config",
                "pyproject.toml",
            )
        )

    # We have more than one rule
    assert mocked_run_checks.call_count > 1

    # We expect one line per occurrence
    assert mocked_write.call_count == mocked_run_checks.call_count



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class AbstractClassesInheritFromAbcRule(Rule):
    """
    Checks if classes having "abstract" in their name inherit from Pythons "abstract base class".
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}006"
    RULE_LABEL = 'Abstract classes have to inherit from "abc.ABC".'

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            if not isinstance(node, ast.ClassDef):
                continue
            if "abstract" in node.name.lower():
                # Check whether the class inherits from `ABC`
                inherits_abc = any(self.is_abc(base=base) for base in node.bases)
                uses_abcmeta = self.has_abcmeta(node=node)

                if not (inherits_abc or uses_abcmeta):
                    occurrences.append(
                        Occurrence(
                            rule_id=self.RULE_ID,
                            rule_label=self.RULE_LABEL,
                            filename=self.filename,
                            identifier=node.name,
                            line_number=node.lineno,
                        )
                    )

        return occurrences

    def is_abc(self, *, base) -> bool:
        """
        Checks whether a given base class is `ABC`.
        """
        if isinstance(base, ast.Name):
            # Direct reference (e.g. `ABC`)
            return base.id == "ABC"
        elif isinstance(base, ast.Attribute):
            # Module attribute (e.g. `abc.ABC`)
            return base.attr == "ABC"
        return False

    def has_abcmeta(self, *, node) -> bool:
        """
        Checks whether the class `ABCMeta` is used as a metaclass.
        """
        for keyword in node.keywords:
            if keyword.arg == "metaclass":
                if isinstance(keyword.value, ast.Name):
                    return keyword.value.id == "ABCMeta"
                elif isinstance(keyword.value, ast.Attribute):
                    return keyword.value.attr == "ABCMeta"
        return False



from boa_restrictor.rules.django.prohibit_assert_raises import AssertRaisesProhibitedRule
from boa_restrictor.rules.python.abstract_class_inherits_from_abc import AbstractClassesInheritFromAbcRule
from boa_restrictor.rules.python.asterisk_required import AsteriskRequiredRule
from boa_restrictor.rules.python.dataclass_kw_only import DataclassWithKwargsOnlyRule
from boa_restrictor.rules.python.global_import_datetime import GlobalImportDatetimeRule
from boa_restrictor.rules.python.return_type_hints import ReturnStatementRequiresTypeHintRule
from boa_restrictor.rules.python.service_class_only_one_public import ServiceClassHasOnlyOnePublicMethodRule

BOA_RESTRICTOR_RULES = (
    AsteriskRequiredRule,
    ReturnStatementRequiresTypeHintRule,
    GlobalImportDatetimeRule,
    DataclassWithKwargsOnlyRule,
    ServiceClassHasOnlyOnePublicMethodRule,
    AbstractClassesInheritFromAbcRule,
)

DJANGO_BOA_RULES = (AssertRaisesProhibitedRule,)


def get_rules(*, use_django_rules: bool) -> tuple:
    """
    Returns a list of all enabled rules.
    """
    if use_django_rules:
        return BOA_RESTRICTOR_RULES + DJANGO_BOA_RULES
    return BOA_RESTRICTOR_RULES



import ast

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.python.return_type_hints import ReturnStatementRequiresTypeHintRule


def test_function_has_return_and_type_hint():
    source_tree = ast.parse("""def my_function(self) -> bool:
            return True
        """)

    occurrences = ReturnStatementRequiresTypeHintRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0


def test_function_has_return_missing_type_hint():
    source_tree = ast.parse("""def my_function(self):
            return True
        """)

    occurrences = ReturnStatementRequiresTypeHintRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        line_number=1,
        rule_id=ReturnStatementRequiresTypeHintRule.RULE_ID,
        rule_label=ReturnStatementRequiresTypeHintRule.RULE_LABEL,
        identifier="my_function",
    )


def test_function_missing_return_has_type_hint():
    source_tree = ast.parse("""def my_function(self) -> bool:
            pass
        """)

    occurrences = ReturnStatementRequiresTypeHintRule.run_check(filename="my_file.py", source_tree=source_tree)

    assert len(occurrences) == 0



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class DataclassWithKwargsOnlyRule(Rule):
    """
    This rule will enforce that you use the `kw_only` parameter in every dataclass decorator.
    This will force the developer to set all dataclass attributes as kwargs instead of args, which is more explicit and
    easier to refactor.
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}004"
    RULE_LABEL = 'Enforces "kw_only" parameter in dataclass decorator.'

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.ClassDef):
                for decorator in node.decorator_list:
                    if (
                        (
                            isinstance(decorator, ast.Call)
                            and (
                                (isinstance(decorator.func, ast.Name) and decorator.func.id == "dataclass")
                                or (isinstance(decorator.func, ast.Attribute) and decorator.func.attr == "dataclass")
                            )
                        )
                        or (isinstance(decorator, ast.Name) and decorator.id == "dataclass")
                        or (isinstance(decorator, ast.Attribute) and decorator.attr == "dataclass")
                    ):
                        # We use the default that "kw_only" is absent
                        kw_only_present = False
                        if isinstance(decorator, ast.Call):
                            kw_only_present = any(
                                isinstance(arg, ast.keyword) and arg.arg == "kw_only" and arg.value.value is True
                                for arg in decorator.keywords
                            )
                        if not kw_only_present:
                            occurrences.append(
                                Occurrence(
                                    filename=self.filename,
                                    rule_label=self.RULE_LABEL,
                                    rule_id=self.RULE_ID,
                                    line_number=node.lineno,
                                    identifier=None,
                                )
                            )

        return occurrences



import dataclasses
from typing import Optional


# TODO: Add kw_only when we drop Python 3.9 support
@dataclasses.dataclass
class Occurrence:
    rule_id: str
    rule_label: str
    filename: str
    identifier: Optional[str]
    line_number: int
